
# Jaesub Hong (jhong@cfa.harvard.edu)

if 'clise' in __file__:
	from clise.jsontool import cc
else:
	from jsontool import cc

import pandas as pd
# import numpy as np
import astropy
import subprocess

from astropy.io	 import fits, ascii
from collections   import OrderedDict
from astropy.table import Table, QTable
from os		 import path
# from colorama      import Fore
from statistics	 import mean, median
from IPython	 import embed
from datetime      import datetime
from pathlib       import Path

import re
import gzip
# import csv

from bs4 import BeautifulSoup

def to_num(candidate):
	"""parse string to number if possible
	work equally well with negative and positive numbers, integers and floats.

	Args:
		candidate (str): string to convert

	Returns:
		float | int | None: float or int if possible otherwise None
	"""
	try:
		float_value = float(candidate)
	except ValueError:
		return None

	# optional part if you prefer int to float when decimal part is 0
	if float_value.is_integer():
		return int(float_value)

	return float_value

scino = re.compile(r'[+\-]?(?:0|[1-9]\d*)(?:\.\d*)?(?:[eE][+\-]?\d+)?')
def get_val_err(array, strip=False):
	# need to vectorize this....
	# and implement this in tabletool to separate them out
	ans  = []
	err  = []
	lerr = []
	# if type(array[0]) is not str: return array
	doerr  = False
	dolerr = False
	for each in array:
		# val=re.split('(\D|^\.)',each)
		# val=re.search('^\s*([-\+][0-9.]+)',each)
		if strip: each = each.strip().replace(" ", "")
		val = re.findall(scino, each)
		if   len(val) == 0:
			ans.append(0.0)
			err.append(0.0)
			lerr.append(0.0)
			print('cannot find a number in', each)
		elif len(val) == 1:
			ans.append(float(val[0]))
			err.append(0.0)
			lerr.append(0.0)
		elif len(val) == 2:
			ans.append(float(val[0]))
			err.append(float(val[1]))
			lerr.append(-float(val[1]))
			# lerr.append(0.0)
			doerr = True
		elif len(val) >= 3:
			ans.append(float(val[0]))
			e1 = float(val[1])
			e2 = float(val[2])
			if e2 < 0:
				lerr.append(e2)
				err.append(e1)
			else:
				lerr.append(e1)
				err.append(e2)
			dolerr = True
			if len(val) > 3:
				print(each, val)
	if not doerr:   err = None
	if not dolerr: lerr = None
	return ans, err, lerr
# ----------------------------------------------------------------------------
# pandas to astro
def pd_to_ast(table):
	ans = Table.from_pandas(table)
	for each in table:
		stype = str(table[each].dtype)
		if bool(re.match('Float', stype)):
			ans[each] = ans[each].astype(float)
		elif bool(re.match('Int', stype)):
			ans[each] = ans[each].astype(int)
	return ans

# astro to panda
def ast_to_pd(data, nopandas=False):
	data.convert_bytestring_to_unicode()
	if nopandas:
		# import plottool as pt
		# pt.embed()
		return data
	try:
		return data.to_pandas()
	except ValueError:
		# this is when some columns are multi-dimensional
		return data

# ----------------------------------------------------------------------------
# write the table to a fits file; handle gzip compression
def to_fits(filename, table, overwrite=True, allstring=False):
	filename = str(Path(filename).expanduser())
	ft = table
	if type(table) is pd.core.frame.DataFrame:
		ft = pd_to_ast(table)
		if allstring:
			for key in ft.keys(): ft[key] = ft[key].astype('str')

	if overwrite: ft.write(filename, overwrite=overwrite, format='fits')
	else:
		uz_filename = filename
		if bool(re.search(r'\.gz$', filename)):
			uz_filename = re.sub(r'\.gz$', '', filename)

		if uz_filename != filename:
			if path.exists(filename):
				subprocess.check_output(['gzip -f -d ' + filename], shell=True).decode()

		ft.write(uz_filename, append=True, format='fits')

		if uz_filename != filename:
			subprocess.check_output(['gzip -f  ' + uz_filename], shell=True).decode()

# write the table to a csv or fits file
def to_csv_or_fits(filename, table, overwrite=True, index=False, allstring=False):
	filename = str(Path(filename).expanduser())
	basename = path.basename(filename)
	if   bool(re.search(r'\.csv(|\.gz)$', basename)):
		if type(table) is astropy.table.table.Table: table = Table.to_pandas(table)
		table.to_csv (filename, index=index)
	elif bool(re.search(r'\.fits(|\.gz)$', basename)):
		to_fits(filename, table, overwrite=overwrite, allstring=allstring)

def read_mca(filename, columns=None):
	if columns is None: columns = ['data']
	with open(filename, encoding='utf-8', errors='ignore') as f:
		lines = f.readlines()

	data = []
	start = False
	for line in lines:
		if bool(re.search(r'<<END>>',  line)): break
		if start:
			data.append(int(line))
		if bool(re.search(r'<<DATA>>', line)): start = True

	return pd.DataFrame(data, columns=columns)

# read the table from a csv or fits file or a few other formats
def from_csv_or_fits(filename, index=False, ftype=None, hdu=1, withHeader=False,
					 nopandas=False, fill=-1, columns=None, comment=None,
					 dropna=False):
	filename = str(Path(filename).expanduser())
	basename = path.basename(filename)

	if ftype is not None:
		if   ftype == "json": return pd.read_json(filename)
		elif ftype == "csv" : 
			ans = pd.read_csv (filename, index_col=index, comment=None)
			if dropna: ans = ans.dropna()
			return ans
		elif ftype == "mca" : return read_mca (filename, columns=columns)
		# elif ftype == "tiff": return read_img (filename)
		elif ftype == "fits": 
			ans = ast_to_pd(Table.read(filename, format='fits', hdu=hdu), nopandas)
			if dropna: ans = ans.dropna()
			if withHeader: 
				return ans, fits.open(filename)[hdu].header
			return ans
		elif ftype == "fitshdr": return fits.open(filename)[hdu].header
		else:
			try:    return ast_to_pd(Table.read(filename, format=ftype), nopandas)
			except Exception: pass

	if   bool(re.search( r'\.csv(|\.gz)$', basename)): 
		ans = pd.read_csv (filename, index_col=index, comment=comment)
		if dropna: ans = ans.dropna()
		return ans
	elif bool(re.search( r'\.mrt(|\.gz)$', basename)): 
		ans = ast_to_pd(Table.read(filename, format='ascii.mrt'),     nopandas)
		if dropna: ans = ans.dropna()
		return ans
	elif bool(re.search( r'\.mca(|\.gz)$', basename)): return read_mca    (filename, columns=columns)
	# elif bool(re.search(r'\.tiff(|\.gz)$', basename)): return read_img    (filename)
	elif bool(re.search( r'\.xml(|\.gz)$', basename)): return pd.read_xml (filename)
	elif bool(re.search(r'\.json(|\.gz)$', basename)): return pd.read_json(filename)
	# elif bool(re.search(r'\.fits(|\.gz)$', basename)): return ast_to_pd(Table.read(filename, format='fits', hdu=hdu), nopandas)
	else: 
		try:
			ans = Table.read(filename, format='fits', hdu=hdu)
		except:
			import plottool as pt 
			pt.embed()
		ans = ast_to_pd(ans, nopandas)
		if dropna: ans = ans.dropna()
		if withHeader: 
			return ans, fits.open(filename)[hdu].header
		return ans

# read fits as astropy
def readfits(infile, hdu=1, hold=True):
	infile = str(Path(infile).expanduser())
	data = Table.read(infile, format='fits', hdu=hdu)

	if hold: embed()
	return data

# read SAORT output
def from_saort(infile):
	infile = str(Path(infile).expanduser())
	data   = ascii.read(infile, delimiter=r'\s', comment='^;')
	header = data.meta['comments']

	mat = re.match(r'^(.+)\((.*)\)_for_(.+)\((.*)\)', header[0])

	ytit   = mat.group(1)
	yunit  = mat.group(2)
	if yunit == 'arcseconds': yunit = 'arcsec'

	coltit  = mat.group(3)
	colunit = mat.group(4)

	cols = header[1].split('\t')

	mat   = re.match(r'^(.+)\((.*)\)', cols[0])
	xtit  = mat.group(1)
	xunit = mat.group(2)

	pdt = []
	for idx, each in enumerate(cols[1:]):
		pdt.append(QTable([data['col1'], data['col' + str(idx + 2)]],
			names=[xtit, ytit], units=[xunit, yunit],
			meta={'extname': coltit + ':' + cols[idx + 1].strip() + colunit,
				'time': str(datetime.now())}))

	return pdt

def to_json(filename, cfg):
	import json
	file = open(path.expanduser(filename), "w")
	cfg_str = json.dumps(cfg)
	file.write(cfg_str.replace(',', ',\n\t'))
	file.close()

# ----------------------------------------------------------------------------
# list the properties of fits tables
def lsfits(infile, hdu=None, overview=True, full=False,
		column=None, header=None, stat=None,
		hold=False):
	"""list the properties of fits tables
	% lsfits filename [-full] [-column HDU] [-header HDU] [-stat HDU:column]
	e.g., lsfits example.fits """

	if infile is None:
		print("Need an input file. Try with -help")
		return

	hdul = fits.open(infile)
	if overview and column is None and header is None and stat is None:
		hdul.info()
	if full or column is not None:
		for each in hdul:
			if each.is_image:  continue
			if type(column) is str:
				if each.name != column: continue
			print(cc.key + each.name, cc.reset)
			print(each.columns)
			for subeach in each.columns:
				print(subeach)
				print(subeach.name)
				print(type(subeach))
	if header is not None:
		for each in hdul:
			if type(header) is str:
				if each.name != header: continue
			print(cc.key + each.name, cc.reset)
			hdr = str(each.header)
			for i in range(0, len(hdr), 80):
				line = hdr[i:i + 79]
				if line != ''.ljust(79): print(line)
				i = i + 80
	if stat is not None:
		hdu, column = stat.split(':')
		data = from_csv_or_fits(infile, index=False, ftype='fits', hdu=hdu)
		print(infile, hdu, column, min(data[column]), max(data[column]))

	if hold:
		from IPython import embed
		embed()
	hdul.close()

# list table columns
def list(infile, index=None, ftype=None, hdu=1, list=None, hold=False,
		maxrow=0, match=None, stat=None, wrap=False, verbose: int = 0):
	"""list table vertically, works for fits and csv tables
	% tab2list filename"""

	inp = from_csv_or_fits(infile, index=index, ftype=ftype, hdu=hdu)

	trows = len(inp)
	if match is not None:
		match = match.split(';')
		for each in match:
			each = each.strip()
			if verbose >= 2: print(each)
			if   '==' in each:
				column, value = each.split('==')
				if inp[column].dtypes != 'object': value = to_num(value)
				inp = inp[inp[column] == value]
			elif '!=' in each:
				column, value = each.split('!=')
				if inp[column].dtypes != 'object': value = to_num(value)
				inp = inp[inp[column] != value]
			elif '~=' in each:
				column, value = each.split('~=')
				if inp[column].dtypes != 'object': value = to_num(value)
				inp = inp[inp[column].str.each(value)]
			elif '>=' in each:
				column, value = each.split('>=')
				if inp[column].dtypes != 'object': value = to_num(value)
				inp = inp[inp[column] >= value]
			elif '<=' in each:
				column, value = each.split('<=')
				if inp[column].dtypes != 'object': value = to_num(value)
				inp = inp[inp[column] <= value]
			elif '>' in each:
				column, value = each.split('>')
				if inp[column].dtypes != 'object': value = to_num(value)
				inp = inp[inp[column] > value]
			elif '<' in each:
				column, value = each.split('<')
				if inp[column].dtypes != 'object': value = to_num(value)
				inp = inp[inp[column] < value]

	if wrap: sep, rj = '\n', 9
	else:	   sep, rj = ', ', 0
	if stat is not None:
		stat  = stat.split(';')
		nrows = len(inp)
		if nrows == 0: print('No data left')
		else:
			for each in stat:
				each = each.strip()
				if inp[each].dtypes != 'object':
					print('column: '.rjust(rj) + each,
						'rows: '  .rjust(rj) + str(nrows) + ' (' + str(trows) + ')',
						'min: '   .rjust(rj) + '{0:.5}'.format(   min(inp[each])),
						'max: '   .rjust(rj) + '{0:.5}'.format(   max(inp[each])),
						'mean: '  .rjust(rj) + '{0:.5}'.format(  mean(inp[each])),
						'median: '.rjust(rj) + '{0:.5}'.format(median(inp[each])),
						'sum: '   .rjust(rj) + '{0:.5}'.format(   sum(inp[each])),
						sep=sep)
				else:
					print('column: '.rjust(rj) + each,
						'rows: '  .rjust(rj) + str(nrows) + ' (' + str(trows) + ')',
						'min: '   .rjust(rj) + '{0:.20}'.format(   min(inp[each])),
						'max: '   .rjust(rj) + '{0:.20}'.format(   max(inp[each])),
						sep=sep)
		if list is not True: return 0

	max_len = 0
	maxrow = int(maxrow)
	if type(inp) is pd.core.frame.DataFrame:
		# this one only for pandas
		new = True
		for idx, row in inp.iterrows():
			if new:
				new = False
				for key in row.keys():
					clen = len(key)
					if max_len < clen:
						max_len = clen
				max_len += 1

			for key, val in row.items():
				try:
					print(f'{key:>{max_len}} | {val:<}')
				except Exception:
					print(f'{key:>{max_len}} |', val)
			print()
			if maxrow > 0:
				if maxrow <= idx + 1: break
	else:
		# astropy Table
		for key in inp.colnames:
			clen = len(key)
			if max_len < clen:
				max_len = clen
		max_len += 1
		for  row in inp.iterrows():
			for key, val in zip(inp.colnames, row):
				if str(type(val)) == "<class 'numpy.ndarray'>":
					vals = []
					for each in val:
						if type(each) is bool or \
							str(type(each)) == "<class 'numpy.bool_'>":
							if each: vals.append('T')
							else:    vals.append('F')
						else: vals.append(str(each))
					val = ",".join(vals)
				print(key.rjust(max_len), '|', val)
			print()
			if maxrow > 0:
				if maxrow <= idx + 1: break

	return 0

# convert csv to fits and vice versa, based on extension
def csv2fits(infile, outfile, ftype=None, hdu=1):
	"""convert csv files to fits files and vice versa

	% csv2fits csv_file  -outfile fits_file
	% fits2csv fits_file -outfile csv_file [-hdu=HDU]

		-hdu: set HDU id """

	inp = from_csv_or_fits(infile, ftype=ftype, hdu=hdu)
	to_csv_or_fits(outfile, inp)

def tdat2csv(infile, outfile, verbose: int = 0, sort=None, slice=None):

	if verbose >= 2: print(infile)

	names = []
	dtype = OrderedDict()
	skiprows = 0

	# if gz compressed
	if bool(re.search(r'\.gz$', infile)):
		with gzip.open(infile, 'rt') as f:
			for line in f:
				skiprows = skiprows + 1
				mat = re.search(r'^field\[(.*)\][ ]*=[ ]*([a-z]+)[0-9]+', line)
				if bool(mat):
					cdtype = mat.group(2)
					if cdtype == 'char': cdtype = 'str'
					if verbose >= 3:
						print(mat.group(1), ' ', cdtype)
					names.append(mat.group(1))
					dtype[mat.group(1)] = cdtype
				if bool(re.match('^<DATA>$', line)):
					if verbose >= 3: print('done!')
					break
	else:
		with open(infile, 'rt') as f:
			for line in f:
				skiprows = skiprows + 1
				mat = re.search(r'^field\[(.*)\][ ]*=[ ]*([a-z]+)[0-9]+', line)
				if bool(mat):
					cdtype = mat.group(2)
					if cdtype == 'char': cdtype = 'str'
					if verbose >= 3:
						print(mat.group(1), ' ', cdtype)
					names.append(mat.group(1))
					dtype[mat.group(1)] = cdtype
				if bool(re.match('^<DATA>$', line)):
					if verbose >= 3: print('done!')
					break

	print(skiprows)

	if verbose >= 3: print(names)
	data = pd.read_csv(infile, names=names, sep='|',
# 		dtype=dtype,
		engine='c',
		index_col=False,
		na_filter=False,
		encoding='cp1252',  # to avoid failing after nrows=1716511
		skiprows=skiprows)

	if verbose >= 3:
		print(type(data))
		print(type(data['name']))
		print(data['name'])

	data = data[data.name != '<END>']

	if sort is not None:
		data = data.sort_values(by=sort)

	if slice is not None:
		sl = slice
		for key in sl:
			print(key, sl[key])
			data = data[data[key] == sl[key]]

	to_csv_or_fits(outfile, data, allstring=True)

def filter(infile=None, data=None, outfile=None, sl=None, expr=None, ftype=None, hdu=1, sort=None, nopandas=False,
		hold=False, all=False, overwrite=True, verbose=0):
	"""Usage: filter file [-slice options]
		-hdu: set HDU id """

	if data is None:
		data = from_csv_or_fits(infile, ftype=ftype, hdu=hdu, nopandas=nopandas)

	if sl is not None:
		for each, val in sl.items():
			key, op = each.split()
			if verbose >0:
				print(key, op, val)
			# import plottool as pt; pt.embed()
			if   op == '==': data = data[data[key] == val]
			elif op == '!=': data = data[data[key] != val]
			elif op == '~=': data = data[ data[key].str.contains(val)]
			elif op == '^=': data = data[~data[key].str.contains(val)]
			elif op == '>=': data = data[data[key] >= val]
			elif op == '<=': data = data[data[key] <= val]
			elif op == '>' : data = data[data[key] >  val]
			elif op == '<' : data = data[data[key] <  val]
	if expr is not None:
		for key, each in expr.items():
			if verbose >0: 
				print(key, '=', each)
				print(len(data), '->', end='')
			# import plottool as pt; pt.embed()
			if key == 'data': data = eval(each)
			else: locals()[key] = eval(each)
			if verbose >0: print(len(data))


	if sort is not None:
		if   type(data) is pd.core.frame.DataFrame  : data = data.sort_values(by=sort)
		elif type(data) is astropy.table.table.Table: data.sort(sort)

	if hold: embed()
	if outfile is not None: to_csv_or_fits(outfile, data, overwrite=overwrite)

	return data

def valerr(infile, outfile=None, cols=None, ftype=None, hdu=1, sort=None, nopandas=False, strip=False,
		hold=False, all=False, overwrite=True):
	"""Usage: valerr file [-cols column_names]

		-hdu: set HDU id """

	data = from_csv_or_fits(infile, ftype=ftype, hdu=hdu, nopandas=nopandas)

	if cols is not None:
		for col in cols:
			val, err, lerr = get_val_err(data[col], strip=strip)
			data[col] = val
			if err  is not None: data[col + '_err' ] = err
			if lerr is not None: data[col + '_lerr'] = lerr

	if sort is not None:
		if   type(data) is pd.core.frame.DataFrame  : data = data.sort_values(by=sort)
		elif type(data) is astropy.table.table.Table: data.sort(sort)

	if hold: embed()
	if outfile is not None: to_csv_or_fits(outfile, data, overwrite=overwrite)

# ----------------------------------------------------------------------------
def html2csv(infile, outfile=None, header=None, start=0, meta=None,
		overwrite=True, hdu=None, units=None, dtype=None, verbose=0):

	# empty list
	data = []

	# for getting the header from
	# the HTML file
	soup = BeautifulSoup(open(infile, encoding="utf-8-sig"), 'html.parser')

	if header is None:
		list_header = []
		header = soup.find_all("table")[0].find("tr")
		for items in header:
			try: list_header.append(items.get_text().encode('utf-8'))
			except Exception: continue
		start = start + 1
	else:
		if type(header).__name__ == 'list':
			list_header = header
		else:
			list_header = header.keys()
			units  = []
			dtype_ = []
			for key, val in header.items():
				if type(val).__name__ == 'list':
					units.append(val[0])
					dtype_.append(val[1])
				else:
					units.append(val)

			if dtype is None:
				if len(dtype_) != 0:
					dtype = dtype_

	if units is None:
		list_units = []
		units = soup.find_all("table")[0].find("tr")[start:start + 1]
		for items in units:
			try:
				list_units.append(text=items.get_text().encode('utf-8'))
			except Exception:
				continue
		start = start + 1
	else:
		list_units = units

	ncol = len(list_header)
	# for getting the data
	HTML_data = soup.find_all("table")[0].find_all("tr")[start:]
	for element in HTML_data:
		sub_data = []
		k = 0
		for sub_element in element:
			try:
				inp = sub_element.get_text().encode('utf-8')
				if dtype is not None:
					if   dtype[k] == "int"  : inp = int(inp)
					# elif dtype[k] == "long" : inp = long(inp)
					elif dtype[k] == "float": inp = float(inp)
				sub_data.append(inp)
				k = k + 1
			except Exception:
				continue
		if len(sub_data) != ncol: continue
		data.append(sub_data)

	# Storing the data into Pandas
	# DataFrame

	if meta is None: meta = OrderedDict()
	if hdu is not None: meta['EXTNAME'] = hdu

	dataFrame = Table(rows=data, meta=meta,
			names=list_header, units=list_units)

	# Converting Pandas DataFrame
	# into CSV file
	if outfile is not None:
		to_csv_or_fits(outfile, dataFrame, overwrite=overwrite)




def sum_and_diff(x, y, _heron_):
	""" add them up """
	print( f"sum: {x} + {y} = {x + y}")
	print(f"diff: {x} - {y} = {x - y}")
	return x + y, x - y

def square(x):
	""" square it """
	if type(x) is list:
		print("square for list")
		ans = 0
		for each in x:
			print(f"\t{each}^2 = {each*each}")
			ans = ans + each * each
		print(f"\tsquare sum: {ans}")

	else:
		ans = x * x
		print(f"square: {x}^2 = {ans}")
	return ans

def show(x, y, name="receiving"):
	""" show what you got """
	print(name)
	if x is not None: print('\tx', type(x).__name__, x)
	if y is not None: print('\ty', type(y).__name__, y)

def check_file(infile, outfile=None):
	""" check file names"""
	from os import path

	if outfile is not None:
		if path.isfile(outfile):
			print(f"{infile:20} >> {outfile:20} is already present.")
		else:
			print(f"{infile:20} >> {outfile:20} is ok.")
	else:
		print(f"{infile:20} ok?")

	return infile, outfile



# heron dives into a module and execute a function just as
# herons dive into water and catch fish.


import json
# import hjson    as hjson

if 'clise' in __file__:
	import clise.jsontool as jt
	import clise.rabbit   as rb
else:
	import jsontool as jt
	import rabbit   as rb

import re
import sys
# import math
# import glb
import hashlib
# import textwrap3  # we modify one of its routines below: len
import inspect

# import numpy  as np

from os		import path, makedirs
from pathlib	import Path
from socket		import gethostname
from collections  import OrderedDict
from datetime     import datetime
from importlib    import import_module
from IPython	import embed

__version__    = '0.2.3'

cc = jt.cc()

def version():
	return __version__

# handle input parameters
def get_func_parameters(func):
	""" check func input parameters and keyword
	and update them with the given cfg dictionary
	"""

	if hasattr(func, '__code__'):
		keys = func.__code__.co_varnames[:func.__code__.co_argcount][::-1]
	else: return [], {}

	sorter = {j: i for i, j in enumerate(keys[::-1])}
	if func.__defaults__ is not None:
		values = func.__defaults__[::-1]
		kwargs = {i: j for i, j in zip(keys, values)}
		sorted_args = tuple(
			sorted([i for i in keys if i not in kwargs], key=sorter.get)
		)
		sorted_kwargs = {
			i: kwargs[i] for i in sorted(kwargs.keys(), key=sorter.get)
		}
	else:
		sorted_args = keys[::-1]
		sorted_kwargs = OrderedDict()

	# now check if there are list and dict for generic inputs
	params = inspect.signature(func).parameters
	collect = False
	for param in params.values():
		if param.kind == inspect.Parameter.VAR_POSITIONAL:
			collect = True
			break
		if param.kind == inspect.Parameter.VAR_KEYWORD:
			collect = True
			break

	return sorted_args, sorted_kwargs, collect

def set_func_parameters(sorted_args, sorted_kwargs, cfg=None):
	""" check func input parameters and keyword
	and update them with the given cfg dictionary
	"""
	updated_args = []
	for par in sorted_args:
		if par in cfg: updated_args.append(cfg[par])
		else:		   updated_args.append(None)

	if sorted_kwargs is not None:
		updated_kwargs = OrderedDict()
		for par in sorted_kwargs:
			if par in cfg: updated_kwargs[par] = cfg[par]
			else:		   updated_kwargs[par] = sorted_kwargs[par]

	# if you meet this, the function doesn't tell us what it needs
	# e.g., math.sin
	if updated_args == []:
		for par in cfg.get('-pars', []):
			# print(par)
			if par in cfg.keys() - updated_kwargs.keys():
				updated_args.append(cfg[par])
	for par in cfg.get('-kwpars', []):
		if par in cfg:
			updated_kwargs[par] = cfg[par]
		else:
			updated_kwargs[par] = None

	# only when -pars are set
	if cfg.get('-collect', False):
		for par in cfg:
			if par in cfg.get('-discard', []): continue
			# if par[0:1] == '-':		 continue
			if not bool(re.search('^[a-zA-Z_]', par)): continue
			# if par in cfg.get('-pars',[]): continue
			if par in sorted_args: continue
			if par not in updated_kwargs.keys():
				updated_kwargs[par] = cfg[par]

	if len(sorted_args) > 0:
		if sorted_args[0] == 'self':
			sorted_args  = sorted_args[1:]
			updated_args = updated_args[1:]

	return updated_args, updated_kwargs

def set_module_parameters(module, cfg, keys=None):
	""" set module global parameters
	"""
	if cfg is None: return
	if keys is None:
		for key in cfg:			setattr(module, key, cfg[key])
	else:
		for key in keys & cfg.keys(): setattr(module, key, cfg[key])

# ----------------------------------------------------------------------------

def global_diagnostic(cfgs):
	if help_text(cfgs[0].get('-help', False), raw=cfgs[0].get('-rawtext', False)): exit()

	userhelp = str(cfgs[0].get('-Help', False))
	if 'True' in userhelp:
		for each in cfgs: Help_text(each)
		exit()
	elif 'False' not in userhelp:
		Help_text(cfgs[0], userhelp)
		exit()

def diagnostic(cfg):

	global show

	if '-main' not in cfg.keys():
		print('no -main: need the main subroutine to execute')
		if '-show' in cfg.keys(): show(cfg, full=True)
		return True

	show = str(cfg.get('-show', None))
	if 'True' in show:
		print(cc.hl + 'job', cfg['-id'] + ':', cfg['-main'], cc.reset)
		jt.show(cfg)
		return True

	return False

def skip(cfg):

	# input files exist?
	for each in jt.str_to_list(cfg.get('-checkin', [])):
		target = cfg.get(each, '')
		if target is None or target == '': continue
		target = str(Path(target).expanduser())
		if not path.exists(target):
			print(cc.err + 'missing input', target, 'skipping...' + cc.reset)
			return True

	# output files exist?
	# this needs to be combined with -timecheck or -sizecheck
	if not cfg.get('-clobber', False):
		for each in jt.str_to_list(cfg.get('-checkout', [])):
			target = cfg.get(each, '')
			if target is None or target == '': continue
			if path.exists(target):
				if cfg['-verbose'] == 0: target = path.basename(target)
				print(cc.err + 'output', target, 'exists. skipping...', cc.reset)
				return True
	else:
		for each in jt.str_to_list(cfg.get('-after', [])):
			key, reference = each.split(':')
			target = cfg.get(key, '')
			if target is None or target == '': continue
			mt_target    = Path(target).expanduser().stat().st_mtime
			if reference not in mt_refer:
				mt_refer[reference] = Path(reference).expanduser().stat().st_mtime
			if mt_target < mt_refer[reference]:
				print(cc.err + key, target, 'is modified earlier than', reference + '. skipping...', cc.reset)
				return True

		for each in jt.str_to_list(cfg.get('-before', [])):
			key, reference = each.split(':')
			target = cfg.get(key, '')
			if target is None or target == '': continue
			mt_target = Path(target).expanduser().stat().st_mtime
			if reference not in mt_refer:
				mt_refer[reference] = Path(reference).expanduser().stat().st_mtime
			if mt_target > mt_refer[reference]:
				print(cc.err + key, target, 'is later than', reference + '. skipping...', cc.reset)
				return True

	for each in jt.str_to_list(cfg.get('-include', [])):
		key, phrase = each.split(':')
		target = cfg.get(key, None)
		if target is None or target == '': continue
		mat = re.search(phrase, target)
		if not bool(mat):
			print(cc.err + key, target, 'does not have', phrase + '. skipping...', cc.reset)
			return True

	# exclude
	for each in jt.str_to_list(cfg.get('-exclude', [])):
		key, phrase = each.split(':')
		target = cfg.get(key, None)
		if target is None or target == '': continue
		mat = re.search(phrase, target)
		if bool(mat):
			print(cc.err + key, target, 'has', phrase + '. skipping...', cc.reset)
			return True

	return False

def create_dir(trgdir):
	if not path.isdir(trgdir):
		if trgdir != '':
			makedirs(trgdir, exist_ok=True)

# ----------------------------------------------------------------------------
# load module
def load_routine(name, cfg=None, show_trace=False):

	levels  = name.split('.')
	nlevels = len(levels)

	if nlevels == 1:
		if type(__builtins__).__name__ == 'dict': routine = __builtins__[levels[0]]
		else:							routine = getattr(__builtins__, levels[0])
		if show_trace: print(cc.key + 'loading:', routine, type(routine).__name__, cc.reset)
		return routine, None

	# this shouldn't be necessary, but since I don't understand something, here it is
	# if levels[0] == 'clise':
	# 	levels = levels[1:]
	# 	levels[0] = 'clise.' + levels[0]
	# 	nlevels = nlevels - 1

	if levels[0] not in sys.modules:
		module = __import__(levels[0])
	else:
		module = sys.modules[levels[0]]

	# potential class name
	class_name_ = '.'.join(levels[0:-1])
	global class_obj, class_name
	if "-object" in cfg:
		if cfg["-object"] in class_obj:
			if show_trace: print(cc.key + 'loading the requested object:', cfg["-object"], cc.reset)
			object = class_obj[cfg["-object"]]
			routine = getattr(object, levels[-1])
			return routine, module
	else:
		if class_name_ in class_obj:
			if show_trace: print(cc.key + 'loading the existing object:', class_name_, cc.reset)
			object = class_obj[class_name_]
			routine = getattr(object, levels[-1])
			return routine, module

	routine = module
	if show_trace: print(cc.key + 'loading:', routine, type(routine).__name__, cc.reset)
	for idx, each in enumerate(levels[1:]):

		#  this shouldn't be necessary, but since I don't understand something, here it is
		#  or this must be a more generic issue, requiring a better solution
		#  i.e., why does it load "each" from "routine"? instead, it ignores "routine",
		#  and look just for "each"
		if idx == 0 and routine.__name__ == 'clise':
			routine = import_module('clise.' + each)
			if show_trace:
				print(cc.key + 'loading:', idx, routine, type(routine).__name__, cc.reset)

			continue

		try:
			routine = getattr(routine, each)
			# if show_trace: print(cc.key + 'loading got attr', cc.reset)

		except AttributeError:
			routine = import_module(each, routine)
			# if show_trace: print(cc.key + 'loading a module', each, cc.reset)

		if show_trace: print(cc.key + 'loading:', idx, routine, type(routine).__name__, cc.reset)

		if idx == nlevels - 3:  # -3 indicates the next to the last name

			icfg = cfg.get('-init', OrderedDict()).copy()
			icfg['-collect'] = icfg.get('-collect', True)  # by default, collect everything in "-init"
			init_args, init_kwargs = set_func_parameters([], [], cfg=icfg)

			# print(init_args, init_kwargs)
			# try:
			if type(routine).__name__ == 'type':
				# now we are loading a new class or attempting

				class_obj_  = routine(*init_args, **init_kwargs)
				routine     = getattr(class_obj_, levels[-1])

				# if not specified, then use class_name_ for the key of the object
				# so that the same class name will share the same object
				# if there needs to be more than an object for the same class,
				# one has to specify "-object" to separate them
				objkey	= cfg.get("-object", class_name_)

				# if you are here, then safe to assign these
				class_name[objkey] = class_name_
				class_obj [objkey] = class_obj_
			# except Exception:
			else:
				# an issue with "try: .. except:" is if exception occurs due to an error in a class
				# one ends up here as well, which will fail
				# oops, not class
				routine = getattr(routine, levels[-1])
			break

	return routine, module

def execute_routine(routine, args, kwargs, cfg=None, show_out=False):
	out = routine(*args, **kwargs)
	if type(out).__name__ == 'function':
		args,  kwargs, collect = get_func_parameters(out)
		cfg['-collect'] = cfg.get('-collect', collect)
		args,  kwargs = set_func_parameters(args, kwargs, cfg=cfg)
		out = out(*args, **kwargs)
	if show_out: print(out)
	return out

# should we use a class
routines   = OrderedDict()
modules    = OrderedDict()
taskkey    = 0
maxtask    = 1
class_obj  = OrderedDict()
class_name = OrderedDict()
output     = OrderedDict()
mt_refer   = OrderedDict()
show	     = ''

# execute configuration file

# to avoid conflict with expression in eval, most variables start and end with _
def inherit(_cfg_):
	global output
	_output_ = output.copy()
	if "-inherit" in _cfg_:
		for _each_ in _output_:
			locals()[_each_] = _output_[_each_]

		# this is not necessarily a scalar value, but it's just the first
		# element of the dictionary values instead of the whole dictionary.
		if "-scalar" in _cfg_["-inherit"]:
			_selected_ = _cfg_["-inherit"]["-scalar"]
			for _each_ in _selected_:
				locals()[_each_] = eval("list(" + _each_ + ".values())[0]")

		# this returns the dictionary values as a list
		if "-list" in _cfg_["-inherit"]:
			_selected_ = _cfg_["-inherit"]["-list"]
			for _each_ in _selected_:
				locals()[_each_] = eval("list(" + _each_ + ".values())")

		for _key_, _val_ in _cfg_["-inherit"].items():
			if _key_[0] == "-": continue
			_cfg_[_key_] = eval(_val_)
	return _cfg_

def execute(cfg):
	""" execute the main routine called in cfg
	"""

	global show

	# rabbit=cfg.get('-rabbit', False) >= True
	block = 'rabbit' in cfg.get('-block', '')
	if diagnostic(cfg):
		if not block: return True
		return False

	job_str = ['job', f"{cfg['-jobc']}/{cfg['-maxjob']}",
		"%-8s: %-25s:" % (cfg['-nid'], cfg['-id']), cfg['-main']]
	job_str = " ".join(job_str)
	if 'job' in show:
		# idf = cfg['-id'].split(':')
		print(cc.hl + job_str, cc.reset)

	if 'input' in show:
		jt.show(cfg, full=False)
		if not block: return True

	if 'hidden' in show:
		jt.show(cfg, full=True)
		if not block: return True

	if 'file' in show:
		jt.show_files(cfg)

	if cfg.get('-dump', '') != '':
		try:
			cfg_str = json.dumps(cfg.copy())
			hfile = open(path.expanduser(cfg['-dump']), "a")
			hfile.write('// ' + job_str + '\n')
			hfile.write(cfg_str.replace(',', ',\n\t'))
			hfile.write(',\n')
			hfile.close()
		except Exception:
			print(cc.err + "cannot dump cfg since serializing cfg failed" + cc.reset)

	# anything not at the base level, assume it's a function inside a class
	# cfg['-init'] = cfg.get('-init',len(cfg['-main'].split('.'))>1)

	global taskkey
	routine, module = load_routine(cfg['-main'], cfg=cfg, show_trace='trace' in show)

	routines[taskkey] = routine
	modules [taskkey] = module

	routine_args,  routine_kwargs, collect = get_func_parameters(routine)
	cfg['-collect'] = cfg.get('-collect', collect)

	if 'func' in show:
		jt.show_feed(routine_args, routine_kwargs, cfg['-main'])
		return True

	# -------------------------------------------------------------------------
	# variable inherit
	cfg = inherit(cfg)
	# -------------------------------------------------------------------------
	# when the routine has clear input parameters exposed, take advantage of it
	# if hasattr(routine,'__code__'):
	# 	routine_keys = routine.__code__.co_varnames[:routine.__code__.co_argcount][::-1]
	# else: routine_keys = []
	#
	# cfg['-pars'] = cfg.get('-pars', routine_keys[::-1])
	# actualy this part requires switching the unique parameter sets
	# default=OrderedDict()
	# for each in cfg['-pars']:
	# 	if each in cfg: default[each] = cfg[each]
	#
	# if len(default)==0: default=copy(cfg)
	# -------------------------------------------------------------------------
	# set module level global parameters
	if '-global' in cfg.keys():
		gvar = cfg['-global']
		# doing this here for safety
		if type(gvar) is list:
			gvar = OrderedDict()
			for each in cfg['-global']:
				if each in cfg.keys(): gvar[each] = cfg[each]
		set_module_parameters(module, gvar)

	if 'time' in show: start_time = datetime.now()

	# if this is a native routine, feed the whole thing either by class itself or cfg
	sorted_args   = OrderedDict()
	sorted_kwargs = OrderedDict()
	sorted_args, sorted_kwargs = set_func_parameters(routine_args, routine_kwargs, cfg=cfg)

	# -------------------------------------------------------------------------
	# feed all the configuration parameters if the input keyword is correct
	roundup = cfg.get('-roundup', '_heron_')
	if   roundup in routine_args : sorted_args  [routine_args.index(roundup)] = cfg
	elif roundup in sorted_kwargs: sorted_kwargs[roundup] = cfg

	# a bit too risky, so not now
	# roundup=cfg.get('-roundupall', '__heron__')
	# if   roundup in routine_args : sorted_args  [routine_args.index(roundup)] = all
	# elif roundup in sorted_kwargs: sorted_kwargs[roundup] = all

	# -------------------------------------------------------------------------
	if 'feed' in show:
		jt.show_feed(routine_args, routine_kwargs, cfg['-main'],
					inargs=sorted_args, inkwargs=sorted_kwargs)

	for each in cfg.get('-mkoutdir', []):
		target = path.dirname(cfg.get(each, ''))
		if not path.exists(target): create_dir(target)

	if skip(cfg): return -1

	# -------------------------------------------------------------------------
	output_ = execute_routine(routine, sorted_args, sorted_kwargs, cfg=cfg,
				show_out='output' in show)

	if 'time' in show:
		runtime(start_time, datetime.now(), cfg['-main'], jobid='job ' + str(taskkey), cc=cc)

	if '-return' in cfg:
		global output
		if type(output_) is not tuple: output_ = [output_]
		for idx, each in enumerate(cfg.get('-return', '').replace(' ', '').split(',')):
			if each not in output:
				output[each] = OrderedDict()
			# this needs a mod for rabbit's task generation
			output[each][cfg['-nid']] = output_[idx]
		return output
	else:
		return output_

	""" """

def go(cfgs=None, befores=None, afters=None, ovw=None, cli=sys.argv[1:], level=-1):
	global taskkey, mt_refer, maxtask, show

	if cfgs is None: cfgs, befores, afters, ovw = jt.load_cli(cli)
	if type(cfgs) is not list: cfgs = [cfgs]

	global_diagnostic(cfgs)

	# search for self containing rabbit and reconfig?
	# cfgs=jt.reconfig_to_alltaskgen(cfgs)

	level = level + 1
	if level == 0:
		taskkey  = 0
		mt_refer = OrderedDict()
		maxtask  = len(cfgs)

	start_time = datetime.now()

	for idx, each in enumerate(cfgs):
		taskkey = taskkey + 1

		each['-level' ] = level
		each['-jobc'  ] = taskkey
		each['-maxjob'] = maxtask

		# this doesn't work
		# need to set this in jsontool 
		# and then execute here? or execute as a special job
		# or use another script?
		# if "-restart" in each:
		# 	# clean up existing class object
		# 	print("cleaning up existing class object")
		# 	global class_obj, class_name
		# 	class_name = OrderedDict()
		# 	import plottool as pt
		# 	for key, obj in class_obj.items():
		# 		print(key, obj)
		# 		# pt.embed()
		# 		del obj
		# 		# pt.embed()
		# 	class_obj = OrderedDict()

		out = execute(each)

		if type(befores) is list: before = befores[idx]
		else:				  before = befores
		if type(afters ) is list: after  = afters [idx]
		else:				  after  = afters

		_, after_ = rb.get_common_vs_match(after)
		after_    = jt.merge(after_, ovw)

		# if each.get('-rabbit', False) >=True and type(out) is not bool:
		if 'rabbit' in each.get('-block', '') and type(out) is not bool:
			# id augmentation by level depth
			# a bit adhoc: too intertwined with internal steps of jt.load_cli
			# need a more clean break btw the two
			if len(out) == 0: continue

			maxtask = maxtask + len(out)

			for idx2, each2 in enumerate(out):
				# update the number and full id
				each2['-id'], each2['-nid'], each2['-fid'] \
						= jt.set_id(each2.get('-id', ''), idx2 + 1, pref=each['-nid'] + '.')

				# only applied selected cases in after?
				# other wise this supercede
				# what if one needs to apply globally?
				out[idx2] = jt.get_task_ready(each2, before, after_, doprocess=level + 2)

			go(out, befores=before, afters=after, level=level)

	global show
	end_time = datetime.now()
	if 'time' in show:
		runtime(start_time, end_time, '', jobid='job all', cc=cc)
	log(cfgs, start_time, end_time)

# ----------------------------------------------------------------------------
def log(cfg, start_time, end_time):
	""" write a log
	"""
	# called at the end of the program
	# requires the directory name: set by _hisdir
	# the filename is automatically assigned by date_time,
	# but can be set manually by _hisfile

	if cfg[0].get('-dryrun', False):     return
	if cfg[0].get('-logfile', '') == "": return

	if not cfg[0].get('-save_cli_only_run', False):
		if jt.json_counter == 0: return

	hostname = gethostname()

	try:
		cfg_str = json.dumps(cfg)
	except Exception:
		print(cc.err + "cannot log this since serializing cfg failed" + cc.reset)
		return
	hashres = hashlib.sha1(cfg_str.encode())

	# ------
	for each in cfg:
		if '-main' in each:
			main_name = each['-main'][0]
			break
	lmain     = len(cfg)
	if lmain > 0: main_name = main_name + '+' + str(lmain)
	loginfo = ['{}'.format(start_time)[2:19], '{}'.format(end_time - start_time),
			'SHA1:' + hashres.hexdigest(),
			main_name, '', hostname,
			'']
	loginfo = ' '.join(loginfo)

	logfile = path.expanduser(cfg[0]['-logfile'])
	if path.isfile(logfile):
		lfile = open(logfile, "a")
		lfile.write('\n' + loginfo)
	else:
		lfile = open(logfile, "w")
		lfile.write(loginfo)
	lfile.close()

	# ------
	hisfile = cfg[0].get('-hisfile', None)
	if hisfile is None: return
	if hisfile == '_auto_':
		filename = start_time.strftime("%y%m%d_%H%M%S") + '_' + main_name + '.json5'
		hisfile = (Path(logfile).parent / Path(filename)).expanduser()

	hfile = open(path.expanduser(hisfile), "w")
	hfile.write('//' + loginfo + '\n')
	hfile.write(cfg_str.replace(',', ',\n\t'))
	hfile.close()

	""" """

def runtime(start_time, end_time, label, jobid='', cc=cc):
	print(cc.key + jobid + ':', '{}'.format(end_time - start_time),
			cc.type + '{}'.format(start_time),
			cc.key + label, cc.reset)

# ----------------------------------------------------------------------------
def help_text(display=False, raw=False):
	""" print out the help text of cjpy and cjson
	"""

	display  = str(display)
	if 'True' in display:
		print("""Usage: clise JSON_input_file1 -options_for_file1 ...
				[JSON_input_file2 -options_for_file2 ...]
				[--with common_JSON_files -common_options ...]
				[--WITH common_JSON_files -common_options ...]

		clise --help [keys]
		clise [JSON_files ...] --Help
		clise --main module.routine --Help """.replace("\n\t\t", "\n"))
		return True
	if 'False' in display: return False

	contents = OrderedDict()
	contents['heading'    ] = '### CLISe:'
	contents['contents'   ] = '### Table of Contents'
	contents['overview'   ] = '### Quick overview of the basic concept'
	contents['install'    ] = '### Installation and startup'
	contents['features'   ] = '### Features and limitation'
	contents['features'   ] = '### Features in parameter setting with JSON files'
	contents['iteration'  ] = '### Iterative calling and file search'
	contents['Sequential' ] = '### Sequential calling'
	contents['cli'	    ] = '### Options for command line parameters'
	contents['parameters' ] = '### Parameter list'
	contents['logging'    ] = '### Logging'
	contents['more'       ] = '### More'
	contents['changes'    ] = '### Changes'
	contents['end'        ] = ''
	keys = list(contents.keys())

	dir = path.dirname(__file__)
	rf = open(dir + '/doc/heron.md', "r")
	readme = rf.read()
	rf.close()

	if display == "all": display = keys[:-1]
	else:		         display = display.split(',')

	from rich.console import Console
	from rich.markdown import Markdown
	console = Console()

	for each in display:
		if each == "keys":
			print(keys[:-1] + ['all'])
			continue
		if each in keys:
			idx = keys.index(each)
			phrase = '(' + contents[each] + '.+)' + contents[keys[idx + 1]]
			mat = re.search(phrase, readme, re.DOTALL)
			if not bool(mat): continue
			selected = mat.group(1)
			if raw: print(selected)
			else:   console.print(Markdown(selected))

	return True
	""" """

# ----------------------------------------------------------------------------
def Help_text(cfg, what=None, keyword='-main', cc=cc):
	"""
	print out the Help text from the called routines
	No argument:
		if a single routine is called, then print out __doc__ of the routine
		if multiple routines are called, list the modules and functions called
	with an argument for a routine:
		the print out __doc__ of the routine whose name is the same as argument
		this enables printing out any doc function
	with an argument for a module:
		list all the functions of the module
		this argument ends with .
	"""

	mdict, mlist = jt.extract(cfg, key=keyword)
	jt.show(mdict, full=True)
	if len(mlist) == 1:
		routine, module = load_routine(mlist[0], cfg=cfg)
		if routine.__doc__ is None: print('hmm, does not have any doc in', mlist[0])
		else:				    print(routine.__doc__.replace("\n\t", "\n"))
	else:
		jt.show(mdict, full=True)

	if what is None: return True

	# forgot what's the point of all these.....
	if what in mlist:
		print(cc.key + what + cc.reset)
		routine, module = load_routine(what, cfg=cfg)
		if routine.__doc__ is None: print('hmm, does not have any doc in', what)
		else:				    print(routine.__doc__.replace("\n\t", "\n"))
	else:
		if what[-1] == '.':
			what = what[:-1]
			print('searching functions in module:', cc.key + what + cc.reset)

			from inspect import isfunction
			whats = what.split('.')
			if len(whats) > 1:
				routine, module = load_routine(what, cfg=cfg)
			else:
				routine = __import__(what)
			members = dir(routine)
			for each in members:
				if isfunction(getattr(routine, each)): print('	', each)
		else:
			print(cc.err + what, 'is not called yet' + cc.reset)
			try:
				routine, module = load_routine(what, cfg=cfg)
				if routine.__doc__ is None: print('hmm, does not have any doc in', what)
				else:				    print(routine.__doc__.replace("\n\t", "\n"))
			except Exception:
				print(cc.err + "cannot load", what + cc.reset)

	return True
	""" """

def main():
	go()
# ----------------------------------------------------------------------------
if __name__ == "__main__":
	"""
	in bash, use this program in this way
	alias heron='python heron.py'

	and then
	heron json_file [json_files ...] [-other parameters]

	to run in ipython

	import heron as hr
	cfg=jt.load_cli('cli_text')
	out=hr.go(cfg)

	"""

	go()
	""" """




import re
# import json
# import pyjson5 as json5
# import hjson    as hjson

if 'clise' in __file__:
	import clise.jsontool as jt
else:
	import jsontool as jt

# import sys
# import math
import glob
# import hashlib
# import textwrap3  # we modify one of its routines below: len

# import numpy  as np

from os		import path, chdir, getcwd, makedirs
from pathlib	import Path
# from socket		import gethostname
from collections  import OrderedDict
from datetime     import datetime
# from importlib    import import_module
# from IPython	import embed

cc = jt.cc()
# ----------------------------------------------------------------------------
# grab time
def timestr2stamp(tstr):

	match = re.search(r'([0-9]+)([^0-9])[0-9]+([^0-9])[0-9]+', tstr)
	if match is None:
		try:    modtime = Path(tstr).expanduser().stat().st_mtime
		except Exception: return None
		return modtime

	date = tstr.split()
	if len(date) == 1:
		time = None
		date = date[0]
	elif len(date) >= 2:
		time = date[1]
		date = date[0]
	else:
		return None

	match = re.search(r'([0-9]+)([^0-9])[0-9]+([^0-9])[0-9]+', date)
	if match is None: return None

	if len(match.group(1)) <= 2: yr = '%y'
	else:				     yr = '%Y'
	sep1 = match.group(2)
	sep2 = match.group(3)

	dformat = yr + sep1 + "%m" + sep2 + "%d"
	if time is None:
		return datetime.timestamp(datetime.strptime(date, dformat))

	match = re.search(r'[0-9]+([^0-9])[0-9]+([^0-9])[0-9]+', time)

	sep3 = match.group(1)
	sep4 = match.group(2)

	tformat = "%H" + sep3 + "%M" + sep4 + "%S"
	return datetime.timestamp(datetime.strptime(tstr, dformat + " " + tformat))

def sizestr2size(sstr):
	try:
		resize = int(sstr)
	except Exception:
		try: resize = Path(sstr).expanduser().stat().st_size
		except ValueError: resize = -1
	return resize

# fix column type from panda table to astropy table
# perhaps a better way to do this, but for now...

def new_ext(filename, new):
	return re.sub(r'\.[^\.]+$', '.' + new, filename)
# -------------------------------------------------------------------
# semi-regex handling in the parameters regarding file names
def find_tag(expression, inp, tagid='', pre='{', post='}'):
	""" get tags from a parameter
	"""
	if expression is None: return ''

	# this means inp can be drived from expression itself by removing '(' and ')'
	if inp is None:
		inp = expression.replace('(', '')
		inp = inp.replace(')', '')
		return_inp = True
	else:
		return_inp = False

	subexpr = re.findall(expression, inp)
	if type(subexpr[0]) is tuple: subexpr = [item for t in subexpr for item in t]

	tag = OrderedDict()

	for idx, expr in enumerate(subexpr):
		tag[pre + tagid + str(idx + 1) + post] = expr

	if return_inp:  return tag, inp
	else: 	    return tag
	""" """

def apply_tag(inp, tags):
	""" apply tags to a parameter
	"""
	if tags is None: return inp
	if inp  is None: return inp

	out = inp
	for key in tags: out = out.replace(key, tags[key])

	return out
	""" """

def combine_tags(keys, tags, ids=None, cfg=None):
	""" get more tags from other files used in the program
	"""

	if cfg  is None: cfg = OrderedDict()
	if ids  is None: ids = [key + ':' for key in keys]

	if tags is None: tags = OrderedDict()
	combined = tags.copy()

	values = []
	for ckey, cid in zip(keys, ids):
		expre		= apply_tag(cfg[ckey], tags)
		tag, value  = find_tag(expre, None, tagid=cid)
		values.append(value)
		for key_, tag_ in tag.items(): combined[key_] = tag_

	return combined, values
	""" """

# -------------------------------------------------------------------
def regex2searchable(regex, wild='*'):
	""" convert regular expression to file searchable expression
	"""
	if not bool(re.search(r'\(', regex)): return regex

	spexpr = regex.split('(')

	newexpr = []
	for seg in spexpr:
		if not bool(re.search(r'\)', seg)):
			newexpr.append(seg)
			continue

		newexpr.append(re.sub(r'(.*\))', '*', seg))

	return ''.join(newexpr)
	""" """

def mixed2regexsrchable(mixed, wild='*', single='?'):
	""" convert mixed expression to file searchable and regex expression
	"""
	if not bool(re.search(r'\(', mixed)):
		regex = mixed
		regex = regex.replace('.', '\.')
		regex = regex.replace(wild, '.*')
		regex = regex.replace(single, '.')
		return mixed, regex

	spexpr = mixed.split('(')

	srchable = []
	regex = []

	for seg in spexpr:
		if not bool(re.search(r'\)', seg)):
			reg = seg
			reg = seg.replace('.', '\.')
			reg = reg.replace(wild, '.*')
			reg = reg.replace(single, '.')
			regex.append(reg)
			srchable.append(seg)
			continue

		srchable.append(re.sub(r'.*\)', '*', seg))

		mat = re.match(r'(.*\))(.*)', seg)
		reg = mat.group(2)
		reg = reg.replace('.', '\.')
		reg = reg.replace(wild, '.*')
		reg = reg.replace(single, '.')
		regex.append(mat.group(1) + reg)

	srchable = ''.join(srchable)
	srchable = srchable.replace('**', '*')
	regex = '('.join(regex)
	return srchable, regex
	""" """

# ----------------------------------------------------------------------------
# filtering file selections
def pick(infile, indir=None, include=None, exclude=None, verbose=0):
	""" see if the given filename meets the requirement as an input file
	"""
	# set by "from" and/or "except" options
	if infile == "": return False

	pick = True
	if include is not None: pick = pick and     bool(re.search(include, infile))
	if exclude is not None: pick = pick and not bool(re.search(exclude, infile))

	if not pick:
		if verbose >= 3: print('not picking', infile)
		return False

	if indir is not None: infile_ = (Path(indir) / Path(infile)).expanduser()
	else:                 infile_ = infile

	if not path.isfile(infile_):
		if verbose >= 2: print(infile, 'does not exist.')
		return False

	return pick
	""" """

# iteration loop settings
def search_infile(infile_expre, tagid='', indir=None,
		include=None, exclude=None,
		recursive_search=True, sortby='', onlylist=False,
		verbose=0):
	""" make file list from wild card input file parameters
	"""

	cdir = getcwd()
	if indir is not None: chdir(path.expanduser(indir))

	infile_expre = jt.str_to_list(infile_expre)

	# convert regex to searchable expression
	# e.g.,  dummy_([0-9]+).fits => dummy_*.fits
	canfiles = OrderedDict()

	prev_fullname = None
	for idx, expre in enumerate(infile_expre):

		# this allows tagging but somewhat too cryptic and unconventional
		# avoid id starting with a number since number id is reserved for automatic assignment
		mat = re.search(r'^(.*)\s+\>([^0-9].*)$', expre)

		if bool(mat): fullname, id = mat.group(1, 2)
		else:		  fullname, id = expre, str(idx)

		if fullname == "==":
			if prev_fullname is None:
				print(cc.err + 'requires at least one valid file name expression' + cc.reset)
				exit()
			else:
				fullname = prev_fullname

		dir  = path.dirname (fullname)
		file = path.basename(fullname)

		if dir == '.': dir = ''
		if dir != '' : dir = dir + '/'

		schdir,  rexdir  = mixed2regexsrchable(dir)
		schfile, rexfile = mixed2regexsrchable(file)

		x = OrderedDict()
		x['schdir' ] = schdir
		x['schfile'] = schfile
		x['rexfile'] = rexdir + rexfile
		x['isregex'] = schfile != file

		canfiles[id] = x

		prev_fullname = fullname

	# now actual file search for each searchable file expression
	ans = OrderedDict()
	names = []
	for id, cfile in canfiles.items():

		if cfile['schfile'] == '': continue

		# not sure why the search is this convoluted
		if indir is not None:
			if recursive_search: files = glob.glob('**/' + cfile['schfile'], recursive=recursive_search)
			else:			   files = glob.glob(cfile['schfile'])
		else:				   files = glob.glob(cfile['schdir'] + cfile['schfile'], recursive=recursive_search)

		if len(files) == 0: continue

		files.sort(key=lambda x:[int(c) if c.isdigit() else c for c in re.split(r'(\d+)', x)])

		# check if the found match with the original regular expression
		if cfile['isregex']:
			survived = []
			tags     = []
			for found in files:
				if not re.search(cfile['rexfile'], found): continue
				survived.append(found)
				tags.     append(find_tag(cfile['rexfile'], found, tagid=tagid))
		else:
			survived = files
			tags	   = [OrderedDict()] * len(files)

		if len(survived) == 0: continue

		# if infile_order != 'unsort':
		# 	zipped = zip(survived, tags)
		# 	sortee = sorted(zipped, reverse= infile_order == 'reverse')
		# 	tuples = zip(*sortee )
		# 	survived, tags = [ list(tuple) for tuple in tuples]

		# collect any that survives
		for file, tag in zip(survived, tags):
			x = OrderedDict()
			x['name'] = file
			x['tag' ] = tag
			x['id'  ] = id
			x['pick'] = pick(file, indir=indir, include=include, exclude=exclude, verbose=verbose)
			key = file + ' ' + str(id)
			# unique key is determined by the filename and group id
			ans[key] = x
			names.append(file)

	chdir(cdir)

	if onlylist: return names
	return ans

def get_outfile(outfile, infile, outdir=None, outsubdir=None,
		mirror_indir=True, tags=None, switch_subdir=None,
		cfg=None, verbose=0):
	""" get output file name based on input filename and other options
	"""

	if outfile is None: return None, None

	if cfg     is None: cfg = OrderedDict()
	if outdir  is None: mirror_indir = False

	insubdir = path.dirname(infile)
	if outsubdir is None:
		outsubdir = insubdir
		if outsubdir != '': outsubdir = outsubdir + '/'

	if switch_subdir is True:
		for src, trg in switch_subdir.items():
			outsubdir = outsubdir.replace(src, trg)

	if tags is not None:
		outsubdir = apply_tag(outsubdir, tags)
		outfile   = apply_tag(outfile,   tags)

	if mirror_indir:
		outdir  = (Path(outdir) / Path(outsubdir)).expanduser()
		outfile = (Path(outdir) / Path(outfile)  ).expanduser()
	else:
		# usually in this case, this function is not likely needed
		if outdir is None:
			outdir  = Path(path.dirname(outfile)).expanduser()
		else:
			outfile = (Path(outdir) / Path(outfile)).expanduser()

	return outfile, outdir
	""" """

# ----------------------------------------------------------------------------
# this part can be combined with matchID and get_task_Ready in jsontool
# essentially the same thing, but it is done in a slightly different way
regtk = ['==', '!=', '~=', '~!=']
regex = '(' + "|".join(regtk) + ")"

def group_cfg_for_match_(seed, matkey=''):
	common = OrderedDict()
	common['-id'] = ''
	common['-main'] = seed['-main']

	global regtk, regex

	matchc = OrderedDict()
	for each in regtk: matchc[each] = OrderedDict()

	for key in seed:
		if key == matkey: continue
		mat = re.search('^' + matkey + regex + '(.*)', key)
		if not bool(mat):
			common[key] = seed[key]
			continue
		con, par = mat.group(1, 2)
		for each in regtk:
			if con == each:
				matchc[each][key] = seed[key]
	return common, matchc

def update_cfg_by_match_(task, seed, matchc, matkey=None, matval=None):

	# now inherit this specific options
	# more general to specific
	for key in matchc['!=']:
		mat  = re.search('^' + matkey + '!=(.*)', key)
		par  = mat.group(1)
		mat_ = re.search('(.*):(.*)', par)
		if bool(mat_):
			par, newkey = mat_.group(1, 2)
			if par != matval: task[newkey] = seed[key]
		else:
			if par != matval:
				for each in seed[key]: task[each] = seed[key][each]

	for key in matchc['~!=']:
		mat  = re.search('^' + matkey + '~!=(.*)', key)
		par  = mat.group(1)
		mat_ = re.search('(.*):(.*)', par)
		if bool(mat_):
			par, newkey = mat_.group(1, 2)
			if not bool(re.search(par, matval)):
				task[newkey] = seed[key]
		else:
			if not bool(re.search(par, matval)):
				for each in seed[key]: task[each] = seed[key][each]

	for key in matchc['~=']:
		mat  = re.search('^' + matkey + '~=(.*)', key)
		par  = mat.group(1)
		mat_ = re.search('(.*):(.*)', par)
		if bool(mat_):
			par, newkey = mat_.group(1, 2)
			if bool(re.search(par, matval)):
				task[newkey] = seed[key]
		else:
			if bool(re.search(par, matval)):
				for each in seed[key]: task[each] = seed[key][each]

	for key in matchc['==']:
		mat  = re.search('^' + matkey + '==(.*)', key)
		par  = mat.group(1)
		mat_ = re.search('(.*):(.*)', par)
		if bool(mat_):
			par, newkey = mat_.group(1, 2)
			if par == matval:
				task[newkey] = seed[key]
		else:
			if par == matval:
				for each in seed[key]: task[each] = seed[key][each]

	return task

def group_cfg_for_match(seed):
	global regtk, regex

	common = OrderedDict()
	matchc = OrderedDict()
	for each in regtk: matchc[each] = OrderedDict()

	for key in seed:
		mat = re.search('^(.+)' + regex + '(.+)', key)
		if not bool(mat):
			common[key] = seed[key]
			continue
		con, par = mat.group(2, 3)
		for each in regtk:
			if con == each:
				matchc[each][key] = seed[key]
	return common, matchc

def group_cfg_for_match_simple(seed):
	global regtk, regex

	common = OrderedDict()
	matchc = OrderedDict()

	for key in seed:
		mat = re.search('^(.+)' + regex + '(.+)', key)
		if not bool(mat):
			common[key] = seed[key]
			continue
		matchc[key] = seed[key]
	return common, matchc

def update_cfg_by_match(task, matchc, matkey=None, matval=None):

	# now inherit this specific options
	# more general to specific
	# need to track that no update has been made...
	for key in matchc['!=']:
		mat = re.search('^' + matkey + '!=(.*)', key)
		if not bool(mat): continue
		par  = mat.group(1)
		mat_ = re.search('(.*):(.*)', par)
		if bool(mat_):
			par, newkey = mat_.group(1, 2)
			if par != matval: task[newkey] = matchc['!='][key]
		else:
			if par != matval:
				for each in matchc['!='][key]: task[each] = matchc['!='][key][each]

	for key in matchc['~!=']:
		mat = re.search('^' + matkey + '~!=(.*)', key)
		if not bool(mat): continue
		par  = mat.group(1)
		mat_ = re.search('(.*):(.*)', par)
		if bool(mat_):
			par, newkey = mat_.group(1, 2)
			if not bool(re.search(par, matval)):
				task[newkey] = matchc['~!='][key]
		else:
			if not bool(re.search(par, matval)):
				for each in matchc['~!='][key]: task[each] = matchc['~!='][key][each]

	for key in matchc['~=']:
		mat = re.search('^' + matkey + '~=(.*)', key)
		if not bool(mat): continue
		par  = mat.group(1)
		mat_ = re.search('(.*):(.*)', par)
		if bool(mat_):
			par, newkey = mat_.group(1, 2)
			if bool(re.search(par, matval)):
				task[newkey] = matchc['~='][key]
		else:
			if bool(re.search(par, matval)):
				for each in matchc['~='][key]: task[each] = matchc['~='][key][each]

	for key in matchc['==']:
		mat = re.search('^' + matkey + '==(.*)', key)
		if not bool(mat): continue
		par  = mat.group(1)
		mat_ = re.search('(.*):(.*)', par)
		if bool(mat_):
			par, newkey = mat_.group(1, 2)
			if par == matval:
				task[newkey] = matchc['=='][key]
		else:
			if par == matval:
				for each in matchc['=='][key]: task[each] = matchc['=='][key][each]

	return task

def get_common_vs_match(seed):
	global regtk, regex

	common = OrderedDict()
	matchc = OrderedDict()

	for key in seed:
		mat = re.search('^(.+)' + regex + '(.+)', key)
		if not bool(mat):
			common[key] = seed[key]
			continue
		matchc[key] = seed[key]
	return common, matchc

# ----------------------------------------------------------------------------
# seed  block: common for every iteration of each task
# cases block: includes
# 	common block: common for all tasks for all iteration
# 	matchc block: common for each iteration (and each task)

# this is more compatible with legacy code cjpy, but obsolete
def multiply_by_id(seed, matkey='-id', verbose=0):

	common, matchc = group_cfg_for_match(seed)
	tasks = []

	# cannot use -nid, -fid since they are not known at this time
	# only -id and -main is known
	# the order matters...
	udkey = OrderedDict()
	udkey['-main'] = seed['-main'].split('.')[-1]

	for idval in seed[matkey]:

		# first inherit common ones
		task = common.copy()

		# set proper id
		# task['-id'], task['-nid'], task['-fid'] = jt.set_id(id, idx+1)
		task [matkey] = idval
		udkey[matkey] = idval

		if '-changes' in common:
			from deepdiff import DeepDiff as dd
			before = task.copy()

		# update by matching key
		for ematkey, ematval in udkey.items():
			task = update_cfg_by_match(task, matchc, matkey=ematkey, matval=ematval)

		if '-changes' in common:
			diff = dd(before, task)
			if len(diff) == 0:
				print(cc.key + 'No changes by the -id selection for', idval, cc.reset)

		tasks.append(task)

	return tasks

# this is technically obsolete by the new scheme
def multiply_by_id_multi(seed, cases=None, matkey='-id', verbose=0, **kwpars):

	if type(seed) is str      : seed = jt.load_file(seed)
	if type(seed) is not list : seed = [seed]

	common = OrderedDict()
	for eseed in seed[:]:
		if '-main' not in eseed:
			common = jt.merge(common, eseed)
			seed.remove(eseed)

	if cases is not None: common = jt.merge(common, cases)
	common, matchc = group_cfg_for_match(common)

	tasks = []

	udkey = OrderedDict()
	for idval in common[matkey]:
		for sidx, eseed in enumerate(seed):

			# first inherit common ones
			task = common.copy()
			task = jt.merge(task, eseed)
			task['-id'   ] = idval + "." + str(sidx + 1)

			# the order is meaningful for the exact match ('==')?
			udkey['-main'] = eseed['-main'].split('.')[-1]
			udkey[matkey ] = idval + "." + str(sidx + 1)

			# update by matching key
			for ematkey, ematval in udkey.items():
				task = update_cfg_by_match(task, matchc, matkey=ematkey, matval=ematval)

			tasks.append(task)

	return tasks

# this is more compatible with legacy code cjpy, not recommended
def multiply_by_file(seed, inkey="infile", outkey=None,
				indir=None, outdir=None, outsubdir=None, basedir=None,
				mirror_indir=False, switch_subdir=False,
				checkin=None, checkout=None,
				include=None, exclude=None,
				appkeys=None, tagkeys=None,
				rel2indir=None, rel2outdir=None,
				rel2nodir=None,  # files relative to no set dir
				recursive=True, sortby='', reverse=None,
				before=None, after=None,
				larger=None, smaller=None,
				verbose=0):

	# among rel2indir, rel2outdir, rel2nodir, all the file keys should be covered

	if type(seed) is str: seed = jt.load_file(seed)

	if checkin is None: checkin = [inkey]

	infiles = search_infile(seed[inkey], indir=indir, include=include, exclude=exclude,
						recursive_search=recursive, sortby=sortby,
						verbose=verbose)

	if infiles is None: return []

	common, matchc = group_cfg_for_match(seed)

	tasks = []

	for mainkey, info in infiles.items():

		name = info.get('name' , None)
		tag  = info.get('tag'  , None)
		id   = info.get('id'   , None)
		pick = info.get('pick' , True)

		if name is None: continue
		if not pick: continue

		# first inherit common ones
		task = common.copy()

		task['-id'] = mainkey
		task[inkey] = name

		task = update_cfg_by_match(task, matchc, matkey=inkey,	          matval=name)
		task = update_cfg_by_match(task, matchc, matkey='-' + inkey + '.key', matval=id  )

		# need to update tkeys and akeys
		# search for outfile?
		# incheck outcheck...
		# all these done by set_file_pars?

		# need to fix this
		if tagkeys is not None:
			tag, values = combine_tags(tagkeys, tag, ids=tagids, cfg=seed)
			for tkey, value in zip(tkeys, values): task[tkey]=value

		if appkeys is not None:
			for each in appkeys: task[each] = apply_tag(task[each], tag)

		if outkey in seed:
			outfile = seed.get(outkey, None)
			# print('====', outdir, outfile)
			outfile, outdir_ = get_outfile(outfile, name,
						outdir=outdir, outsubdir=outsubdir,
						mirror_indir=mirror_indir, switch_subdir=switch_subdir,
						tags=tag, cfg=seed, verbose=verbose)
			task[outkey] = Path(outfile).expanduser()
			# print('====', outdir, outfile)

		# expand ~***
		if indir is not None: task[inkey] = (Path(indir) / Path(name)).expanduser()
		else:                 task[inkey] = Path(name).expanduser()

		# update files in a directory relative to input dir
		if rel2indir is not None:
			fulldir = path.dirname(task[inkey])
			for key in rel2indir:
				if key in seed:
					if task[key] is not None:
						task[key] = (Path(fulldir) / Path(task[key])).expanduser()

		# update files in a directory relative to output dir
		if rel2outdir is not None:
			if outkey is not None:
				if mirror_indir: fulldir = path.dirname(Path(task[outkey]).expanduser())
				else:		     fulldir = Path(outdir).expanduser()

				for key in rel2outdir:
					if key in seed:
						if task[key] is not None:
							task[key] = (Path(fulldir) / Path(task[key])).expanduser()

		# rename file path relative to basedir if possible
		if basedir is not None:
			try   : task[inkey] = task[inkey].relative_to(basedir)
			except Exception: pass

			if outkey in task:
				if outkey is not None:
					try   : task[outkey] = task[outkey].relative_to(basedir)
					except Exception: pass

			for each in [rel2indir, rel2outdir, rel2nodir]:
				if each is None: continue
				if type(each) is str: each = [each]
				for key in each:
					if key not in task: continue
					if task[key] is None: continue
					try   : task[key] = task[key].relative_to(basedir)
					except Exception: pass

		# update filename with proper strings
		task[inkey] = str(task[inkey])
		if outkey is not None: task[outkey] = str(task[outkey])
		for each in [rel2indir, rel2outdir, rel2nodir]:
			if each is None: continue
			if type(each) is str: each = [each]
			for key in each:
				if key not in task: continue
				if task[key] is None: continue
				task[key] = str(task[key])

		if '-checkin'  not in task: task['-checkin' ] = []
		if '-checkout' not in task: task['-checkout'] = []

		task['-checkin'] = task['-checkin'] + checkin
		if checkout is not None:
			task['-checkout'] = task['-checkout'] + checkout

		tasks.append(task)
		# from IPython	import embed
		# embed()

	# here we have to sort when we have both infile and outfile
	if before is not None:
		if sortby  == "":   sortby  = inkey + ":mtime"
		if reverse is None: reverse = True
	if after is not None:
		if sortby  == "":   sortby  = inkey + ":mtime"
		if reverse is None: reverse = False

	if larger is not None:
		if sortby  == "":   sortby  = inkey + ":size"
		if reverse is None: reverse = False
	if smaller is not None:
		if sortby  == "":   sortby  = inkey + ":size"
		if reverse is None: reverse = True

	if reverse is None: reverse = False
	if sortby != "":
		key, stype = sortby.split(':')
		for task in tasks:
			if   stype == "mtime"    : task["-sval"] = Path(task[key]).stat().st_mtime
			elif stype == "size"     : task["-sval"] = Path(task[key]).stat().st_size
			elif stype == "name"     : task["-sval"] = path.basename(task[key])
			elif stype == "fullname" : task["-sval"] = task[key]
			else: continue
		tasks = sorted(tasks,  key=lambda x: x['-sval'], reverse=reverse)

	return tasks
	""" """

def multiply_by_file_multi(seed, cases=None, inkey="infile", outkey="outfile",
					indir=None, outdir=None, outsubdir=None, basedir=None,
					mirror_indir=True, switch_subdir=False,
					checkin=None, checkout=None,
					include=None, exclude=None,
					appkeys=None, tagkeys=None,
					rel2indir=None, rel2outdir=None,
					rel2nodir=None,  # files relative to no set dir
					recursive=True, sortby='', reverse=None,
					before=None, after=None,
					larger=None, smaller=None,
					verbose=0):

	# among rel2indir, rel2outdir, rel2nodir, all the file keys should be covered

	if type(seed) is str      : seed = jt.load_file(seed)
	if type(seed) is not list : seed = [seed]

	common = OrderedDict()
	infile_expre = ''
	if checkin is None: checkin = [inkey]

	for idx, eseed in enumerate(seed):
		seed[idx], match = group_cfg_for_match_simple(eseed)
		if '-main' in eseed:
			if infile_expre == '': infile_expre = eseed[inkey]
		else:
			common = jt.merge(common, seed[idx].copy())
			# seed.remove(seed[idx])

		common = jt.merge(common, match)
		if checkout is None:
			if outkey in eseed: checkout = [outkey]

	infiles = search_infile(infile_expre, indir=indir, include=include, exclude=exclude,
						recursive_search=recursive, sortby=sortby,
						verbose=verbose)

	if infiles is None: return []

	if cases is not None: common = jt.merge(common, cases)
	common, matchc = group_cfg_for_match(common)

	tasks = []

	udkey = OrderedDict()
	for mainkey, info in infiles.items():

		name = info.get('name' , None)
		tag  = info.get('tag'  , None)
		id   = info.get('id'   , None)
		pick = info.get('pick' , True)

		if name is None: continue
		if not pick: continue

		for sidx, eseed in enumerate(seed):
			if '-main' not in eseed: continue
			# first inherit common ones
			task = common.copy()
			task = jt.merge(task, eseed)

			task['-id'] = mainkey + "." + str(sidx)  # this should be last?
			task[inkey] = name

			# the order is meaningful for the exact match ('==')?
			udkey['-main'] = eseed['-main'].split('.')[-1]
			udkey['-id'  ] = mainkey + "." + str(sidx)  # this should be last?
			udkey[inkey  ] = name
			udkey['-' + inkey + '.key'] = id

			for ematkey, ematval in udkey.items():
				task = update_cfg_by_match(task, matchc, matkey=ematkey, matval=ematval)

			# need to update tkeys and akeys
			# search for outfile?
			# incheck outcheck...
			# all these done by set_file_pars?

			# need to fix this
			if tagkeys is not None:
				tag, values = combine_tags(tagkeys, tag, ids=tagids, cfg=eseed)
				for tkey, value in zip(tkeys, values): task[tkey] = value

			if appkeys is not None:
				for each in appkeys:
					if each in task: task[each] = apply_tag(task[each], tag)

			if outkey in eseed:
				outfile = eseed[outkey]
				outfile, outdir_ = get_outfile(outfile, name,
							outdir=outdir, outsubdir=outsubdir,
							mirror_indir=mirror_indir, switch_subdir=switch_subdir,
							tags=tag, cfg=eseed, verbose=verbose)
				task[outkey] = Path(outfile).expanduser()

			# expand ~***
			if inkey in task:
				if indir is not None: task[inkey] = (Path(indir) / Path(name)).expanduser()
				else:                 task[inkey] = Path(name).expanduser()

				# update files in a directory relative to input dir
				if rel2indir is not None:
					fulldir = path.dirname(task[inkey])
					for key in rel2indir:
						if key in task:
							if task[key] is not None:
								task[key] = (Path(fulldir) / Path(task[key])).expanduser()

			# update files in a directory relative to output dir
			if outkey in task:
				if rel2outdir is not None:
					if outkey is not None:
						if mirror_indir: fulldir = path.dirname(Path(task[outkey]).expanduser())
						else:		     fulldir = Path(outdir).expanduser()

						for key in rel2outdir:
							if key in task:
								if task[key] is not None:
									task[key] = (Path(fulldir) / Path(task[key])).expanduser()

			# rename file path relative to basedir if possible
			if basedir is not None:
				if inkey in task:
					try   : task[inkey] = task[inkey].relative_to(basedir)
					except Exception: pass

				if outkey in task:
					if outkey is not None:
						try   : task[outkey] = task[outkey].relative_to(basedir)
						except Exception: pass

				for each in [rel2indir, rel2outdir, rel2nodir]:
					if each is None: continue
					if type(each) is str: each = [each]
					for key in each:
						if key not in task: continue
						if task[key] is None: continue
						try   : task[key] = task[key].relative_to(basedir)
						except Exception: pass

			# update filename with proper strings
			if inkey in task: task[inkey] = str(task[inkey])
			if outkey in task:
				if outkey is not None: task[outkey] = str(task[outkey])
			for each in [rel2indir, rel2outdir, rel2nodir]:
				if each is None: continue
				if type(each) is str: each = [each]
				for key in each:
					if key not in task: continue
					if task[key] is None: continue
					task[key] = str(task[key])

			if '-checkin'  not in task: task['-checkin' ] = []
			if '-checkout' not in task: task['-checkout'] = []

			task['-checkin'] = task['-checkin'] + checkin
			if checkout is not None:
				task['-checkout'] = task['-checkout'] + checkout

			tasks.append(task)

	# here we have to sort when we have both infile and outfile
	if before is not None:
		if sortby  == "":   sortby  = inkey + ":mtime"
		if reverse is None: reverse = True
	if after is not None:
		if sortby  == "":   sortby  = inkey + ":mtime"
		if reverse is None: reverse = False

	if larger is not None:
		if sortby  == "":   sortby  = inkey + ":size"
		if reverse is None: reverse = False
	if smaller is not None:
		if sortby  == "":   sortby  = inkey + ":size"
		if reverse is None: reverse = True

	if reverse is None: reverse = False
	if sortby != "":
		key, stype = sortby.split(':')
		for task in tasks:
			if   stype == "mtime"    : task["-sval"] = Path(task[key]).stat().st_mtime
			elif stype == "size"     : task["-sval"] = Path(task[key]).stat().st_size
			elif stype == "name"     : task["-sval"] = path.basename(task[key])
			elif stype == "fullname" : task["-sval"] = task[key]
			else: continue
		tasks = sorted(tasks,  key=lambda x: x['-sval'], reverse=reverse)

	return tasks
	""" """

# ----------------------------------------------------------------------------
def read_multiple_deco(func):
    from functools    import wraps

# turning this into an iterator would be better
    @wraps(func)
    def read_multiple(*args, infile=None, # indir=None,
        hdu=0, tagid='',  id=None,
        include=None, exclude=None,
        start=0, stop=None, appkeys=None,
        recursive_search=True, **kwpars):

        if infile is None: 
            print("No infile. exiting...")
            exit()

        infiles = search_infile(infile, tagid=tagid, # indir=indir,
                    include=include, exclude=exclude,
                    recursive_search=recursive_search, 
                    onlylist=False)

        nfiles = len(infiles) 
        print(nfiles)

        w2r = [] # what to read

        for key in infiles:
            val = infiles[key]
            cfile = val['name']
            new = OrderedDict()
            new['infile'] = cfile
            new['tag'] = val['tag']
            if id is not None:
                new['id'] = apply_tag(id, val['tag'])
            # if outfile is not None:
            #     # not taking full advantage of rabbit.py features
            #     # better way to do this
            #     new['outfile'] = apply_tag(outfile, val['tag'])
            if appkeys is not None:
                  for akey in appkeys:
                        target = kwpars.get(akey,"")
                        new[akey] = apply_tag(target, val['tag'])

		# if it's fits file and require hdu
            if type(hdu) is list:  
                # check all the blocks
                if len(hdu) == 0:
                    hdul = fits.open(cfile)
                    for id, chdu in enumerate(hdul):
                        # skip table
                        htype = type(chdu).__name__
                        if 'Table' in htype: continue

                        # skip empty block
                        if type(chdu.data).__name__ == 'NoneType': 
                            continue
                        new['hdu']  = chdu.name
                        w2r.append(new)
                    hdul.close()
                else:
                    for chdu in hdu:
                        new['hdu']  = chdu
                        w2r.append(new)

                # everything like images
            else:
                new['hdu']  = hdu
                w2r.append(new)

        # if outfile is not None:
        #     kwpars['outfile'] = outfile

        infile=[]
        for idx, each in enumerate(w2r):
            if idx < start: continue
            if stop is not None: 
                if idx > stop: continue
            infile.append(each)

        return func(*args, infile=infile, **kwpars)

    return read_multiple

def read_multiple_deco_wdir(func):
    from functools    import wraps

# turning this into an iterator would be better
    @wraps(func)
    def read_multiple_wdir(*args, infile=None, indir=None,
        hdu=0, tagid='',  id=None,
        include=None, exclude=None,
        start=0, stop=None, appkeys=None,
	  outdir=None, outsubdir=None, outfile=None,
	  mirror_indir=True, switch_subdir=False,
        recursive_search=True, mkoutdir=[], **kwpars):

        if infile is None: 
            print("No infile. exiting...")
            exit()

        infiles = search_infile(infile, tagid=tagid, indir=indir,
                    include=include, exclude=exclude, 
                    recursive_search=recursive_search, 
                    onlylist=False)

        nfiles = len(infiles) 
        # print(nfiles)

        w2r = [] # what to read

        for key in infiles:
            val = infiles[key]
            cfile = val['name']
            new = OrderedDict()
            new['infile'] = cfile
            new['indir'] = indir
            new['tag'] = val['tag']
            if id is not None:
                new['id'] = apply_tag(id, val['tag'])
            # if outfile is not None:
            #     # not taking full advantage of rabbit.py features
            #     # better way to do this
            #     new['outfile'] = apply_tag(outfile, val['tag'])
            if appkeys is not None:
                  for akey in appkeys:
                        target = kwpars.get(akey,"")
                        new[akey] = apply_tag(target, val['tag'])
            # import plottool as pt; pt.embed()
            outfile_, outdir_ = get_outfile(outfile, cfile, 
					outdir=outdir, outsubdir=outsubdir,
					mirror_indir=mirror_indir, switch_subdir=switch_subdir,
					tags=new['tag'])

            new['outfile'] = outfile_
            new['outdir'] = outdir_

            try:
                trgdir = path.dirname(outfile_)
                if not kwpars.get('dry', False):
                      if not path.exists(trgdir): 
                            if 'outfile' in mkoutdir:
                                  if trgdir != '': makedirs(trgdir, exist_ok=True)
            except:
                pass

		# if it's fits file and require hdu
            if type(hdu) is list:  
                # check all the blocks
                if len(hdu) == 0:
                    hdul = fits.open(cfile)
                    for id, chdu in enumerate(hdul):
                        # skip table
                        htype = type(chdu).__name__
                        if 'Table' in htype: continue

                        # skip empty block
                        if type(chdu.data).__name__ == 'NoneType': 
                            continue
                        new['hdu']  = chdu.name
                        w2r.append(new)
                    hdul.close()
                else:
                    for chdu in hdu:
                        new['hdu']  = chdu
                        w2r.append(new)

                # everything like images
            else:
                new['hdu']  = hdu
                w2r.append(new)

        # if outfile is not None:
        #     kwpars['outfile'] = outfile

        infile=[]
        for idx, each in enumerate(w2r):
            if idx < start: continue
            if stop is not None: 
                if idx > stop: continue
            infile.append(each)

        return func(*args, infile=infile, **kwpars)

    return read_multiple_wdir





# Jaesub Hong (jhong@cfa.harvard.edu)

# import csv
# import gzip
import re
import subprocess
# import pandas		as pd
# import astropy.units	as u

# from collections  import OrderedDict
from os		import path, remove
from os		import rename as os_rename
from pathlib	import Path
from datetime	import datetime

if 'clise' in __file__:
	from clise.jsontool import cc, show, simplify
	from clise.rabbit   import timestr2stamp, sizestr2size
else:
	from jsontool import cc, show, simplify
	from rabbit   import timestr2stamp, sizestr2size

skip_job = '_skip_job_'
skip_par = '_skip_par_'

# run command script like HEASOFT or CIAO tools
def cmdscript(cmd, pars, basedir,
		ungzip=[], gzip=[],
		pre=[], post=[],
		skip_job=skip_job, skip_par=skip_par,
		script=True, execute=False, out=False,
		missing=None,
		verbose: int = 0, hold=False):
	"""run generic commands like HEASOFT or CIAO tools
	"""

	skip = False
	cmdscript = []
	if basedir is not None: cmdscript = ["cd " + basedir]

	# pre processing
	if type(pre) is str: pre = [pre]
	for each in pre: cmdscript.append(each)

	# handle if there is no file
	# update gzip list accordingly
	mod_ungzip = [v for v in ungzip]
	mod_gzip   = [v for v in gzip]
	if pars is not None:
		mod_pars = pars.copy()
		for key, val in pars.items():
			if type(val) is not str: continue

			try:
				re.match(skip_par, val)
			except Exception:
				from IPython import embed
				embed()

			if bool(re.match(skip_par, val)):
				mod_pars.pop(key)
			elif bool(re.match(skip_job, val)):
				skip = True
				mod_pars[key] = cc.err + mod_pars[key] + cc.hl
			elif val in ["", "\"\"", "None"]: pass
			else: continue

			if key in ungzip: mod_ungzip.remove(key)
			if key in   gzip: mod_gzip  .remove(key)

	# gzip -d
	for each in mod_ungzip:
		cmdscript.append('gzip -f -d ' + pars[each])
	for each in mod_gzip:
		mod_pars[each] = re.sub(r'\.gz$', '', pars[each])

	display = [v for v in cmdscript]

	# main
	curscript = [cmd]
	if pars is not None:
		for k, v in mod_pars.items():
			# skip parameter
			curscript.append(" " +  k + "=" + str(v))
	cmdscript.append( "".join(curscript))
	display  .append(" \\\n      ".join(curscript))

	# re-gzip
	for each in mod_gzip:
		cmdscript.append('gzip -f ' + mod_pars[each])
		display  .append('gzip -f ' + mod_pars[each])

	# post processing
	if type(post) is str: post = [post]
	for each in post:
		cmdscript.append(each)
		display  .append(each)

	if script :
		if not skip: print("\n".join(display))
		else:        print(cc.hl + "\n".join(display), cc.reset)
		if missing is not None:
			for each in missing:
				print(cc.err + each, cc.reset)

	output = None
	if execute:
		if not skip:
			try:
				output = subprocess.check_output([";".join(cmdscript)], shell=True).decode()
			except subprocess.CalledProcessError as errors:
				print(errors.output)

			if out: print(output)
	else:
		if verbose >= 1: print("No execution requested")

	return output is not None

# recursive search files based on the time and size
def search(infile, refile,
		basedir=None,
		after=None, before=None,
		larger=None, smaller=None,
		hold=False, all=False, feed=False):
	"""Usage: search 'file expr' [-after time|file] [-before time|file]
				[-larger size|file] [-smaller size|file]

	recursively search files based on the time and size
		designed for a command line tool

	e.g.,  search '(.*).txt' -after readme.txt

		lists all the files with the txt extenstion modified
		after readme.txt was lasted modified.
		Note the '' and () in the file expr.
		The wild character should be in () and '.' is needed."""

	infile = str(Path(infile ).expanduser())

	instat = Path(infile).stat()
	intime = instat.st_mtime
	insize = instat.st_size

	cc_time = cc.reset
	cc_size = cc.reset
	cc_name = cc.reset
	cc_sel  = cc.key
	cc_ref  = cc.type

	refile = False

	bingo = True
	compared = False
	if after is not None:
		retime = timestr2stamp(after)
		if retime is not None:
			if retime == intime: refile = True
			if intime <= retime: bingo  = False
			compared = True

	if before is not None:
		retime = timestr2stamp(before)
		if retime is not None:
			if retime == intime: refile = True
			if intime >= retime: bingo  = False
			compared = True

	if smaller is not None:
		resize = sizestr2size(smaller)
		if resize < 0 or insize == resize: refile = True
		if insize >= resize: bingo = False
		compared = True

	if larger is not None:
		resize = sizestr2size(larger)
		if resize < 0 or insize == resize: refile = True
		if insize <= resize: bingo = False
		compared = True

	if bingo:
		if after  is not None or before  is not None: cc_time = cc_sel
		if larger is not None or smaller is not None: cc_size = cc_sel

	if refile:
		cc_size = cc_ref
		cc_time = cc_ref
		cc_name = cc_ref

	if all or bingo or (not compared):
		if feed: print(infile)
		else:    print(cc_time + str(datetime.fromtimestamp(intime))[2:19:],
					cc_size + simplify(insize).rjust(6),
					cc_name + infile + cc.reset)

	return bingo

def search_heron(infile, refile,
		after=None, before=None,
		larger=None, smaller=None,
		hold=False, all=False, feed=False):
	"""Usage: search 'file expr' [-after time|file] [-before time|file]
			[-larger size|file] [-smaller size|file]

	recursively search files based on the time and size
		designed for a command line tool

	e.g.,  search '(.*).txt' -after readme.txt

		lists all the files with the txt extenstion modified
		after readme.txt was lasted modified.
		Note the '' and () in the file expr.
		The wild character should be in () and '.' is needed."""

	infile = str(Path(infile ).expanduser())

	instat = Path(infile).stat()
	intime = instat.st_mtime
	insize = instat.st_size

	cc_time = cc.reset
	cc_size = cc.reset
	cc_name = cc.reset
	cc_sel  = cc.key
	cc_ref  = cc.type

	refile = False

	bingo = True
	compared = False
	if after is not None:
		retime = timestr2stamp(after)
		if retime is not None:
			if retime == intime: refile = True
			if intime <= retime: bingo  = False
			compared = True

	if before is not None:
		retime = timestr2stamp(before)
		if retime is not None:
			if retime == intime: refile = True
			if intime >= retime: bingo  = False
			compared = True

	if smaller is not None:
		resize = sizestr2size(smaller)
		if resize < 0 or insize == resize: refile = True
		if insize >= resize: bingo = False
		compared = True

	if larger is not None:
		resize = sizestr2size(larger)
		if resize < 0 or insize == resize: refile = True
		if insize <= resize: bingo = False
		compared = True

	if bingo:
		if after  is not None or before  is not None: cc_time = cc_sel
		if larger is not None or smaller is not None: cc_size = cc_sel

	if refile:
		cc_size = cc_ref
		cc_time = cc_ref
		cc_name = cc_ref

	if all or bingo or (not compared):
		if feed: print(infile)
		else:    print(cc_time + str(datetime.fromtimestamp(intime))[2:19:],
					cc_size + simplify(insize).rjust(6),
					cc_name + infile + cc.reset)

	return bingo

def rename(infile, outfile, verbose: int = 1, overwrite=False):
	"""Usage: rename infile -outfile outfile [-overwrite]
	rename multiple files using cjson.py

		infile   : only needs a partial expression of the file name
		outfile  : needs the full expression
		overwrite: overwrite even if the file exists

	e.g.,  rename '(.*)(h)(ere)' -outfile '{1}{2}...{3}.txt'
		will rename files:
		here.txt   -> h...ere.txt
		there.txt  -> th...ere.txt

	each {#} corresponds to the text in () in order.
	wild keys are based on rudimentary regex: e.g.,
		(.*)    : 0 or more characters
		(.)     : 1 character
		([0-9]+): 1 or more numbers
	"""
	infile  = str(Path(infile ).expanduser())
	outfile = str(Path(outfile).expanduser())
	if verbose >= 1: print(infile, '->', cc.key + outfile + cc.err, end='')

	if path.isdir(outfile):
		if verbose >= 1: print('...'.rjust(5), outfile, 'is a directory: skip' + cc.reset)
		return False

	if not path.isfile(outfile):
		os_rename(infile, outfile)
		print(cc.reset)
		return True
	else:
		if overwrite:
			print(cc.hl + '...'.rjust(5), outfile, ' exists, but now overwritten by', infile + cc.reset)
			os_rename(infile, outfile)
			return True
		else:
			if verbose >= 1: print('...'.rjust(5), outfile, 'exists: skip' + cc.reset)
			return False

def summary(success, summary=False, overview=False):
	if success is None: return
	if overview:
		show(success, notype=True)
	if summary:
		# from IPython import embed; embed()
		try:    success = success.values()
		except: pass
		# print(sum([v for v in success if v != None]),'out of',len(success))
		print(sum(success), 'out of', len(success))

def delete(infile, verbose=1):
	"""delete infile
	"""
	infile = str(Path(infile).expanduser())
	if path.isfile(infile):
		if verbose >= 1: print('deleting', infile)
		remove(infile)
